/**
  Raytracing demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2010-01-22 (Public Domain)
*/
#include <iostream>
#include <cmath>
#include <GL/glew.h> /* 
OpenGL Extensions Wrangler:  for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <GL/glut.h> /* OpenGL Utilities Toolkit, for GUI tools */
#include "ogl/glsl.h"
#include "ogl/glsl.cpp"
#include "ogl/minicam.h"
#include "osl/mat4.h"
#include "osl/mat4_inverse.h"

#include <vector>

// A quadric object
struct quadric_object {
	osl::mat4 A; // quadric surface matrix: 0 = x^T A x
	
	vec3 color; // diffuse color
	float reflect; // 0=matte; 1=totally specular reflective
	
	osl::mat4 MVinv; // from world to trim space
	
	float refract; // 0=matte; 1=totally transparent
	float refract_index; // refractive index (if refractive)
	float checkerboard; // 0=uniform color; 1=black and white checkers
	float slice; // slices per meter
	
	//float opacity; // 0=transparent; 1=opaque
};

void upload_modelview(osl::mat4 &A,quadric_object &q) 
{
	// Object matrix
	osl::mat4 mv; // object -> world space
	glGetFloatv(GL_MODELVIEW_MATRIX,&mv[0][0]);
	osl::mat4 B=inverse(mv); // world -> object
	q.MVinv=B;
	osl::mat4 BAB=transpose(B)*A*B; // see rule to modify quadric object
	q.A=BAB; // upload quadric matrix
}

void display(void) 
{
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE); // turn on Z writes
	glEnable(GL_BLEND);
	
	glClearColor(0.4,0.5,0.7,0.0); // background color
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); // flush any ancient matrices
	gluPerspective(70.0, // fov
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT),
		0.1,
		2000.0); // z clipping planes
	oglCameraLookAt(1.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity(); // flush any ancient matrices
	
	/* The proxy sphere should be gently rotating */
	const float rotate_rate=0.01; /* revolutions per second */
	float angle=glutGet(GLUT_ELAPSED_TIME)*0.001*rotate_rate*360;
	
	/* Build the pixel shader */
	static GLhandleARB progP=makeProgramObjectFromFiles(
		"raytrace_vertex.txt","raytrace.txt");
	glUseProgramObjectARB(progP);
	glFastUniform3fv(progP,"C",1,camera);


	// Make 2D texture
	int width=glutGet(GLUT_WINDOW_WIDTH), height=glutGet(GLUT_WINDOW_HEIGHT);
	static GLuint texname=0;
	if (texname==0) {
		glActiveTexture(GL_TEXTURE0);
		glGenTextures(1,&texname);
		glBindTexture(GL_TEXTURE_2D,texname);
		
		glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA32F_ARB, // internal format
			width,height,0, // width x height
			GL_RGBA,GL_FLOAT,0);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	}
	
	// Render into texname:
	static GLuint fbname=0;
	if (fbname==0) {
		glGenFramebuffers(1,&fbname);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbname);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
			GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, texname, 0);
		GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
                if (status!=GL_FRAMEBUFFER_COMPLETE_EXT) std::cout<<"Framebuffer:OH NOEES!\n";
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
	}
	glFastUniform1i(progP,"quadriclist",0); // texture unit 0

#if 0
/* Upload checkerboard texture, to texture unit 1 */
	static bool hasChecker=false;
	if (!hasChecker) { /* only need to build and upload one time; it stays set up. */
		hasChecker=true;
		glActiveTexture(GL_TEXTURE1);
		int checkw=128,checkh=128;
		float *checkarr=new float[checkw*checkh];
		for (int y=0;y<checkh;y++) 
		for (int x=0;x<checkw;x++) {
			float cv=0.1;
			if ((x<checkw/2) ^ (y<checkh/2)) cv=0.9;
			checkarr[x+y*checkw]=cv;
		}
		gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGBA8,
			checkw,checkh,
			GL_LUMINANCE,GL_FLOAT,checkarr);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		delete[] checkarr;
	}
	glFastUniform1i(progP,"checkertex",1); // texture unit 1
#endif
	
	// Only draw one set of faces, to avoid drawing raytraced geometry twice.
	//   (front side may get clipped if you're inside the object, so draw back)
	glEnable(GL_CULL_FACE); glCullFace(GL_FRONT); 
	
	glDisable(GL_BLEND); glDisable(GL_ALPHA_TEST); // using alpha as a float
	glDisable(GL_DEPTH_TEST); // no depth buffer

/* Offscreen rendering */
	// Render an image of P into fbname, which writes into texname:
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbname); // draw to offscreen buffer
	glViewport(0,0,width,height);
	glClear(GL_COLOR_BUFFER_BIT); 
	
	// Draw one *HUGE* sphere; this one proxy geometry covers all our objects.
	//   Keep in mind the far clipping plane in determining the sphere radius
	glutSolidSphere(1000.0,3,3); 
	glUseProgramObjectARB(0); // turn off first pass program
	
	
/* Screen rendering pass */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0); // draw to screen again
	glViewport(0,0,width,height); 
	
	// Draw the intermediate buffer onscreen
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glDisable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);
	glBindTexture(GL_TEXTURE_2D,texname);
	glGenerateMipmapEXT(GL_TEXTURE_2D);
	
	static GLhandleARB progS=makeProgramObjectFromFiles(
		"post_vertex.txt","post.txt"
	);
	glUseProgramObjectARB(progS);
	glFastUniform1i(progS,"Ptex",0); // texture unit 0
	glFastUniform3fv(progS,"C",1,camera);	
	
	glEnable(GL_TEXTURE_2D);
	glColor3f(1,1,1);
	glBegin(GL_QUADS);
	glVertex2f(-1,-1);
	glVertex2f(+1,-1);
	glVertex2f(+1,+1);
	glVertex2f(-1,+1);
	glEnd();
	
	glUseProgramObjectARB(0);
	
	
	glutPostRedisplay(); // continual animation
	glutSwapBuffers();

/* Estimate framerate */
	static double last_time=0.0;
	static int framecount=0, last_framecount=0;
	framecount++;
	double cur_time=0.001*glutGet(GLUT_ELAPSED_TIME);
	if (cur_time>0.5+last_time) 
	{ // every half a second, estimate fps
		double time=cur_time-last_time;
		double time_per_frame=time/(framecount-last_framecount);
		char str[100];
		sprintf(str,"IFS Raytracer: %.1f fps, %.1f ms/frame",
			1.0/time_per_frame,1.0e3*time_per_frame);
		glutSetWindowTitle(str);
		last_framecount=framecount;
		last_time=cur_time;
	}
}

int main(int argc,char *argv[]) 
{
	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_RGBA + GLUT_DEPTH + GLUT_DOUBLE);
	glutInitWindowSize(800,600);
	glutCreateWindow("GPU Raytracer");
	
	glewInit();
	
	glutDisplayFunc(display);
	oglCameraInit();
	camera=vec3(0,-3,1);
	
	glutMainLoop();
	return 0;
}
